/**
  *  \file client/chart/tilecontainer.cc
  *  \brief Starchart Tile Container
  *
  *  (c) 2001-2020 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/chart/tilecontainer.h"
#include "client/chart.h"
#include "client/wtile.h"
#include "gfx/gfxcontext.h"
#include "gfx/palette.h"
#include "ui/uiglobal.h"

/*************************** WStarchartTileSkin **************************/

void
WStarchartTileSkin::drawBackground(GfxContext& /*ctx*/, GfxRect /*r*/)
{
    // FIXME: empty?
}

void
WStarchartTileSkin::drawFancyBar(GfxCanvas& can, GfxRect r)
{
    const int icolor = standard_colors[COLOR_SHIELD+3];
    const int fcolor = standard_colors[COLOR_SHIELD+5];
    const int z = standard_colors[COLOR_BLACK];

    int alpha;

    if (ui_screen_bits <= 8) {
        can.drawBar(r, icolor, z, GRAY50_FILL, GFX_ALPHA_OPAQUE);
        alpha = GFX_ALPHA_OPAQUE;
    } else {
        can.drawBar(r, icolor, z, SOLID_FILL, 128);
        alpha = (3*GFX_ALPHA_OPAQUE + GFX_ALPHA_TRANSPARENT)/4;
    }

    can.drawBar(GfxRect(r.x-1,   r.y-2,   r.w+2, 2),   fcolor, z, SOLID_FILL, alpha);
    can.drawBar(GfxRect(r.x-1,   r.y+r.h, r.w+2, 2),   fcolor, z, SOLID_FILL, alpha);
    can.drawBar(GfxRect(r.x-2,   r.y,     2,     r.h), fcolor, z, SOLID_FILL, alpha);
    can.drawBar(GfxRect(r.x+r.w, r.y,     2,     r.h), fcolor, z, SOLID_FILL, alpha);
    can.drawPixel(GfxPoint(r.x-2,     r.y-1),   fcolor, alpha);
    can.drawPixel(GfxPoint(r.x+r.w+1, r.y-1),   fcolor, alpha);
    can.drawPixel(GfxPoint(r.x-2,     r.y+r.h), fcolor, alpha);
    can.drawPixel(GfxPoint(r.x+r.w+1, r.y+r.h), fcolor, alpha);
}

/************************ WStarchartTileContainer ************************/

WStarchartTileContainer::WStarchartTileContainer()
{
    setSkin(skin);
}

void
WStarchartTileContainer::drawContent(GfxCanvas& canvas)
{
    skin.drawFancyBar(canvas, getExtent());
    canvas.update(getExtent());
}

void
WStarchartTileContainer::drawEverything(GfxCanvas& canvas)
{
    drawContent(canvas);
    for (UIBaseWidget* p = getFirstChild(); p != 0; p = p->getNextSibling())
        p->drawEverything(canvas);
}

void
WStarchartTileContainer::drawBelow(UIBaseWidget& /*child*/, GfxCanvas& /*canvas*/)
{
    // draw everything below child, bases override this
}

GfxCanvas&
WStarchartTileContainer::getCanvasFor(UIBaseWidget& /*child*/, ObjectHolder& /*mgr*/)
{
    // get canvas to draw child on, bases override this
    return *ui_screen;
}

bool
WStarchartTileContainer::handleEvent(const UIEvent& /*event*/, bool /*second_pass*/)
{
    // can be overridden by whoever needs it, base provides default
    return false;
}

void
WStarchartTileContainer::onAddRemoveChild(UIBaseWidget& /*the_widget*/, bool /*added*/)
{
    /* FIXME: at first, we assume no uncontrolled add/remove.
       Otherwise, we'd simply have to request relayout.  */
}

void
WStarchartTileContainer::getLayoutInfo(LayoutInfo& li)
{
    li.min_size = li.pref_size = GfxPoint(0, 0);
    li.flags = li_NoLayout;
}

void
WStarchartTileContainer::onResize(GfxRect /*oldsize*/, bool /*by_layout*/)
{
    // Nothing to do?
}

void
WStarchartTileContainer::onChildResize(UIBaseWidget& /*child*/, GfxRect /*oldsize*/)
{
    /* FIXME: at first, we assume no uncontrolled add/remove/resize.
       Otherwise, we'd simply have to request relayout.  */
}

int
WStarchartTileContainer::run()
{
    /* does not run itself */
    return 0;
}

void
WStarchartTileContainer::stop(int /*arg*/)
{ /* does not run itself */ }

void
WStarchartTileContainer::doLayout(GChartViewport& viewport)
{
    enum { EDGE = 5 };

    // Compute total height and width
    int total_height = EDGE;
    int min_width    = 0;
    for (UIBaseWidget* w = getFirstChild(); w != 0; w = w->getNextSibling()) {
        LayoutInfo li;
        w->getLayoutInfo(li);
        total_height += EDGE + li.pref_size.y;
        if (li.pref_size.x > min_width)
            min_width = li.pref_size.x;
    }
    min_width += 2*EDGE;

    // Set own size and allocate space
    GfxRect myExtent(0, 0, min_width, total_height);
    myExtent.moveToEdge(viewport.view, 2, 1, 10);
    setExtent(myExtent, true);

    // Move children
    int y = myExtent.y + EDGE;
    for (UIBaseWidget* w = getFirstChild(); w != 0; w = w->getNextSibling()) {
        LayoutInfo li;
        w->getLayoutInfo(li);
        w->setExtent(GfxRect(myExtent.x + EDGE, y, myExtent.w - 2*EDGE, li.pref_size.y), false);
        y += li.pref_size.y + EDGE;
    }
}

void
WStarchartTileContainer::addNewTile(WTile* w)
{
    addChildWidget(h.add(w), getLastChild());
}

