/**
  *  \file client/widgets/simlist.cc
  *  \brief Class WSimList
  *
  *  (c) 2001-2020 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/widgets/simlist.h"
#include "client/chart.h"       // getPlayerColor
#include "cpluslib/format.h"
#include "game/racename.h"
#include "game/sim-game.h"
#include "game/sim.h"
#include "game/team.h"          // getPlayerId
#include "gfx/gfxcontext.h"
#include "ui/uiglobal.h"
#include "util/nls.h"

static void
setColor(GfxContext& ctx, const GSimObject& obj)
{
    if ((obj.getFlags() & GSimObject::fl_Deactivated) != 0) {
        ctx.setColor(UIColor::tc_Faded);
    } else if (::getPlayerId() != 0) {
        // FIXME: be careful that we don't accidentally pull in the whole map renderer!
        ctx.setColor(getPlayerColor(obj.getOwner()));
    } else {
        // no change, keep default color
    }
}

/******************************** WSimList *******************************/

/** Constructor.
    \param state       [in] Object to display/edit
    \param with_planet [in] Select whether planet is included in list */
WSimList::WSimList(GSimulatorGameInterface& game_if,
                   GSimState& state,
                   bool with_planet)
    : UIListbox(0, 20, 235,
                state.getNumShips() + ((with_planet && state.hasPlanet()) ? 1 : 0)),
      state(state),
      game_if(game_if),
      with_planet(with_planet),
      conn_ship_changed(state.sig_ship_changed.add(this, &WSimList::onShipChanged)),
      conn_planet_changed(state.sig_planet_changed.add(this, &WSimList::onPlanetChanged)),
      conn_structure_changed(state.sig_structure_changed.add(this, &WSimList::onStructureChanged))
{
    updateCount();
}

/** Destructor. */
WSimList::~WSimList()
{ }

void
WSimList::drawPart(GfxCanvas& can, int from, int to)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useFont(*fonts[FONT_NORMAL]);
    for (int i = from; i < to; ++i) {
        const unsigned int index = getTop() + i;
        GfxRect area;
        prepareEntry(ctx, ColorStyle, i, &area);
        if (index < state.getNumShips()) {
            /* ship entry */
            const GSimShip& sh = state.getShip(index);
            setColor(ctx, sh);
            outTextF(ctx, area.x, area.y, area.w, getShipName(sh));
        } else if (index == state.getNumShips() && with_planet && state.hasPlanet()) {
            /* planet entry */
            setColor(ctx, state.getPlanet());
            outTextF(ctx, area.x, area.y, area.w, " " + getPlanetName(game_if, state));
        }
    }
}

void
WSimList::onShipChanged(int /*index*/)
{
    drawWidget();
}

void
WSimList::onPlanetChanged()
{
    drawWidget();
}

void
WSimList::onStructureChanged()
{
    updateCount();
    drawWidget();
}

/** Get currently-selected ship.
    \return ship, or 0 if none selected */
GSimShip*
WSimList::getCurrentShip()
{
    int numShips = state.getNumShips();
    int current  = getCurrent();
    if (current < 0 || current >= numShips) {
        return 0;
    } else {
        return &getSimState().getShip(current);
    }
}

void
WSimList::updateCount()
{
    if (with_planet && state.hasPlanet())
        setSize(state.getNumShips() + 1);
    else
        setSize(state.getNumShips());
    adjustPosition();
    sig_moved.raise(getId());
}

/******************************* Utilities *******************************/

string_t
getShipName(const GSimShip& sh)
{
    const char* fmt;
    if (sh.getFlags() & sh.fl_Deactivated)
        fmt = _(" %d: %s (%s %s, disabled)");
    else
        fmt = _(" %d: %s (%s %s)");

    string_t hull_name;
    if (sh.isCustomShip())
        hull_name = _("custom ship");
    else
        hull_name = getHull(sh.getHull()).getShortName();

    return format(fmt, sh.getId(), sh.getName(), player_racenames.getAdjName(sh.getOwner()), hull_name);
}

string_t
getPlanetName(GSimulatorGameInterface& game_if, const GSimState& st)
{
    if (st.hasPlanet()) {
        string_t name = game_if.getPlanetName(st.getPlanet().getId());
        if (st.getPlanet().getBaseBeamTech() > 0) {
            return format(_("%s (Planet+SB #%d)"), name, st.getPlanet().getId());
        } else {
            return format(_("%s (Planet #%d)"), name, st.getPlanet().getId());
        }
    } else {
        return _("none");
    }
}
