/**
  *  \file flak/vis.h
  *  \brief FLAK: Visualisation
  *
  *  (c) 2001-2020 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_FLAK_VIS_H
#define PCC_V2_FLAK_VIS_H

#include <vector>
#include "client/vcranim.h"
#include "cpluslib/sigslot.h"
#include "flak/flak-db.h"
#include "flak/flak.h"
#include "gfx/gfx.h"
#include "gfx/sprite3d.h"
#include "ui/event.h"
#include "util/idalloc.h"

/** Position on FLAK battlefield. Derived from GfxPoint for
    simplicity. */
struct FlakScreenPos : public GfxPoint {
    int z;

    /** Uninitialized position. */
    FlakScreenPos()
        : GfxPoint() { }
    /** Construct from three coordinate values. */
    FlakScreenPos(int x, int y, int z)
        : GfxPoint(x, y), z(z) { }
    /** Construct from point and Z position. */
    FlakScreenPos(GfxPoint pt, int z)
        : GfxPoint(pt), z(z) { }
};

/** Area on FLAK battlefield. Derived from GfxRect for simplicity. */
struct FlakScreenArea : public GfxRect {
    int z;

    /** Construct empty area at (0,0,0). */
    FlakScreenArea()
        : GfxRect(), z(0) { }
    /** Construct area from 2-D rectangle and Z position. */
    FlakScreenArea(int x, int y, int w, int h, int z)
        : GfxRect(x, y, w, h), z(z) { }
    /** Construct area from 2-D rectangle and Z position. */
    FlakScreenArea(GfxRect r, int z)
        : GfxRect(r), z(z) { }
};

/// Fancy visualizer, base class.
class FlakFancyVisualizer : public FlakVisualizer {
 public:
    /// handleEvent() result.
    enum EventResult {
        eNotHandled,            ///< Event not handled.
        eRedrawMe,              ///< Event handled, redraw battlefield.
        eDontRedrawMe           ///< Event handled, redraw not needed.
    };

    /// Marker "end of fight". Pass this to setTime() to move to the end of the fight.
    static const int32 EndOfFight = 0x7FFFFFFF;

 protected:
    /** Sprite manager used for display. */
    GfxSpriteManager3D manager;

 private:
    /** Number of active effect sprites (beams). As long as effect
        sprites are active, the battle is paused. */
    int num_fx_sprites;

    /** Number of active explosion sprites. Even if the fight is over,
        we have to play until the explosion sprites are gone. */
    int num_explosion_sprites;

    /** Whether any ship was changed and has changed!=0. */
    bool any_ship_changed;

    /** Id association for object sprites. */
    IdAllocator<GfxSprite3D*, uint32> id_assoc;

    /** Whether the current tick() wants us to advance time. */
    bool is_time_advance;

    struct ShipInfo {
        Ptr<GfxSprite3D> box_sprite;
        bool changed : 1;
        ShipInfo()
            : box_sprite(), changed(0) { }
    };
    std::vector<ShipInfo> ship_info;

    /** Stored status. */
    std::vector<FlakStatusToken*> status_tokens;
    /** Status token time divider. */
    int status_time_div;
    /** Maximum number of status tokens. */
    enum { status_max_size = 50 };

    void maybeSaveStatus();
 protected:
    void setObjectSprite(FlakObject& o, Ptr<GfxSprite3D> sprite, bool is_new);
    void markShipChanged(ship_t ship);
 public:
    FlakFancyVisualizer(FlakBattle& battle, GfxRect area);
    ~FlakFancyVisualizer();

    /** Get screen position from world position. */
    virtual FlakScreenPos getScreenPosition(FlakPos fp) = 0;
    /** Get sprite for a ship.
        \param ship ship number
        \return sprite, can be null */
    virtual Ptr<GfxSprite3D> getShipSprite(ship_t ship) = 0;
    /** Get ship from position.
        \param pt screen position
        \return ship id, NO_SHIP if none */
    virtual ship_t getShipAtPosition(GfxPoint pt) = 0;
    /** Handle user event. */
    virtual EventResult handleEvent(const UIEvent& evt) = 0;
    /** Actions to be done every tick. */
    virtual void everyTickAction() = 0;
    /** Create all ship and object sprites.
        \param is_startup true if this is the initialisation of the battle,
        false if this is a repositioning within the fight */
    virtual void createSprites(bool is_startup) = 0;

    void swapStatusTokens(FlakFancyVisualizer& other);

    void addEffectSprite()
        { ++num_fx_sprites; }
    void removeEffectSprite()
        { --num_fx_sprites; }
    void addExplosionSprite()
        { ++num_explosion_sprites; }
    void removeExplosionSprite()
        { --num_explosion_sprites; }

    void init(bool is_startup);
    void doTick(GfxCanvas& can);
    void setShipStatus(ship_t which, bool boxed);
    bool isShipBoxed(ship_t which) const
        { return ship_info[which].box_sprite != 0; }

    void setTime(int32 new_time);

    bool isShipChanged(ship_t ship) const
        { return ship_info[ship].changed; }
    bool isShipChanged() const
        { return any_ship_changed; }

    /** Get sprite associated with some object.
        \param obj the object */
    Ptr<GfxSprite3D> getObjectSprite(FlakObject& obj) const
        {
            if (obj.vis_data == id_assoc.nil)
                return 0;
            else
                return id_assoc.getObject(obj.vis_data);
        }

    void destroyObject(FlakObject& obj, bool violent);
    void destroyShip(ship_t unit);
    void fireBeam(FlakObject& obj1, FlakObject& obj2);
    void fireBeam(FlakObject& obj1, ship_t unit2);
    void fireBeam(ship_t unit1, FlakObject& obj2, bool hits);
    void fireBeam(ship_t unit1, ship_t unit2, bool hits);
    void hitShip(ship_t unit, int expl, int kill, int death_flag);
    // void createObject(FlakObject& obj, ship_t source); must still be implementes

    /** Check whether fight is finished. */
    bool isTerminated() const
        { return the_battle.isTerminated() && num_explosion_sprites == 0; }
    /** Redraw. */
    void draw(GfxCanvas& can)
        { manager.draw(can); }
    /** Updates. */
    void doUpdates(GfxCanvas& can)
        { manager.doUpdates(can); }
    /** Are we in a tick() call which advances time? */
    bool isTimeAdvance() const
        { return is_time_advance; }

    const GfxSpriteManager3D& getManager() const
        { return manager; }
    GfxSpriteManager3D& getManager()
        { return manager; }

    /** Layers (Ids) for sprites. */
    enum {
        layer_Text,             ///< Layer for text (e.g.\ time).
        layer_Boxes,            ///< Layer for boxes (status display).
        layer_Torp,             ///< Layer for torpedoes.
        layer_Beam,             ///< Layer for beams.
        layer_Bang,             ///< Layer for explosions.
        layer_Fighter,          ///< Layer for fighters.
        layer_Ships,            ///< Layer for ships.
        layer_Scene             ///< Layer for scene, e.g.\ stars, grid, etc.
    };

    /** Get layer for an object. */
    static inline int getObjectLayer(FlakObject& o)
        { return o.kind == o.oTorpedo ? layer_Torp : layer_Fighter; }
};

/// Flat visualizer.
class FlakFlatVisualizer : public FlakFancyVisualizer {
    int32 center_x, center_y;
    int32 range;

    std::vector<Ptr<GfxSprite3D> > fleet_sprites;
    std::vector<int> ship_fleets;
    void computeViewport();
 public:
    FlakFlatVisualizer(FlakBattle& battle, GfxRect area);
    ~FlakFlatVisualizer();

    void everyTickAction();
    void createSprites(bool is_startup);

    FlakScreenPos getScreenPosition(FlakPos fp);
    Ptr<GfxSprite3D> getShipSprite(ship_t ship);
    ship_t getShipAtPosition(GfxPoint pt);
    EventResult handleEvent(const UIEvent& evt);

    void createObject(FlakObject& obj, ship_t source);

    int32 getCenterX() const
        { return center_x; }
    int32 getCenterY() const
        { return center_y; }
    int32 getRangeX() const
        { return range; }
    int32 getRangeY() const
        { return range * manager.getBBox().w / manager.getBBox().h; }
};

/// Isometric visualizer.
class FlakIsoVisualizer : public FlakFancyVisualizer {
    int32 center_x, center_y;
    int32 range;

    std::vector<Ptr<GfxSprite3D> > ship_sprites;
    void computeViewport();
 public:
    FlakIsoVisualizer(FlakBattle& battle, GfxRect area);
    ~FlakIsoVisualizer();

    void everyTickAction();
    void createSprites(bool is_startup);

    FlakScreenPos getScreenPosition(FlakPos fp);
    Ptr<GfxSprite3D> getShipSprite(ship_t ship);
    ship_t getShipAtPosition(GfxPoint pt);
    EventResult handleEvent(const UIEvent& evt);

    void createObject(FlakObject& obj, ship_t source);

    int32 getCenterX() const
        { return center_x; }
    int32 getCenterY() const
        { return center_y; }
    int32 getRangeX() const
        { return range; }
    int32 getRangeY() const
        { return range * manager.getBBox().w / manager.getBBox().h; }
};

void playFlakBattle(FlakVcrEntry& battle);

#endif
