/**
  *  \file game/actions/changebuildqueue.h
  *  \brief Class GChangeBuildQueue
  *
  *  (ported from c2ng game::actions::ChangeBuildQueue)
  */
#ifndef PCC_V2_GAME_ACTIONS_CHANGEBUILDQUEUE_H
#define PCC_V2_GAME_ACTIONS_CHANGEBUILDQUEUE_H

#include <vector>
#include "cpluslib/string.h"
#include "cpluslib/types.h"
#include "io/io.h"              // mn32_t (LongProperty_t)
#include "game/global.h"
#include "game/host.h"
#include "game/config.h"
#include "util/random.h"

/** Changing build queue priorities.
    Provides a simple view of the build queue,
    and allows changing priorities (distributing PBx fcodes).

    - create object
    - if desired, call setAvailableBuildPoints()
    - call increasePriority(), decreasePriority(), setPriority() as needed
    - call commit() to write back */
class GChangeBuildQueue {
 public:
    /** Information about a build order in prepared format. */
    struct Info {
        int planetId;                       ///< Planet Id.
        string_t planetName;                ///< Planet name.
        string_t actionName;                ///< Description of build order.
        string_t friendlyCode;              ///< Friendly code.
        int queuePosition;                  ///< Queue position (0 if not known or new order).
        mn32_t pointsRequired;              ///< Number of build points required to build this ship.
        mn32_t pointsAvailable;             ///< Number of build points available for this ship.
        bool hasPriority;                   ///< true if this build order has a priority FC.
        bool conflict;                      ///< true if this priority order conflicts with others (same FC).
        bool playable;                      ///< true if this slot can be modified.

        Info()
            : planetId(), planetName(), actionName(),
              friendlyCode(), queuePosition(), pointsRequired(), pointsAvailable(),
              hasPriority(), conflict(), playable()
            { }
    };

    /** Information about all build orders in prepared format. */
    typedef std::vector<Info> Infos_t;


    /** Constructor.
        \param turn Turn (containing current and previous universe)
        \param host Host version
        \param config Host configuration
        \param rng Random number generator (required to generate fall-back friendly codes)
        \param viewpointPlayer Handle this player's build orders */
    explicit GChangeBuildQueue(GGameTurn& turn,
                              GHost host,
                              const GGameConfig& config,
                              RandomNumberGenerator& rng,
                              int viewpointPlayer);

    /** Destructor. */
    ~GChangeBuildQueue();

    /** Set available build points.
        Call this to populate the "pointsAvailable" field.
        \param points Build points */
    void setAvailableBuildPoints(mn32_t points);

    /** Prepare data into output format.
        Data is reported sorted, highest priority first.
        \param result [out] Data */
    void describe(Infos_t& result) const;

    /** Set priority of a build order.
        \param slot Slot number
        \param pri New priority (1-9 = PBx, 0 = no priority order) */
    void setPriority(size_t slot, int pri);

    /** Increase a slot's priority (build earlier).
        \param slot Slot number */
    void increasePriority(size_t slot);

    /** Decrease a slot's priority (build later).
        \param slot Slot number */
    void decreasePriority(size_t slot);

    /** Write all changes back to universe. */
    void commit();

 private:
    GGameTurn& m_turn;
    // const game::spec::ShipList& m_shipList;
    GHost m_host;
    const GGameConfig& m_config;

    void init(RandomNumberGenerator& rng, int viewpointPlayer);

    void sort();
    void avoid(int setThis, int toThis, size_t slot);

    struct LocalInfo {
        int planetId;               ///< Planet Id.
        int cloningShipId;          ///< Cloning ship Id.
        string_t friendlyCode;      ///< Current friendly code.
        string_t oldFriendlyCode;   ///< Friendly code to revert to (old or random).
        int queuePosition;          ///< Current queue position if known. 0 for new orders.
        bool playable;

        LocalInfo(int planetId, int cloningShipId, string_t friendlyCode, string_t oldFriendlyCode, int queuePosition, bool playable)
            : planetId(planetId), cloningShipId(cloningShipId), friendlyCode(friendlyCode), oldFriendlyCode(oldFriendlyCode), queuePosition(queuePosition), playable(playable)
            { }
    };
    class Sorter;

    std::vector<LocalInfo> m_info;
    mn32_t m_availablePoints;
};

#endif
