/**
  *  \file game/objtypes.h
  *  \brief Object Type
  *
  *  (c) 2001-2020 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GAME_OBJTYPES_H
#define PCC_V2_GAME_OBJTYPES_H

#include "cpluslib/string.h"
#include "cpluslib/sigslot.h"
#include "game/coord.h"

class GUniverse;
class GObject;

/** \class GObjectType
    \brief Object Type

    A class derived from GObjectType defines a set of objects, and iteration through them.

    An object is identified by a non-zero index.
    The index can be checked for validity using isValidIndex().
    A valid index can be turned into an object using getObjectByIndex().

    The base class provides methods getNextIndex() and getPreviousIndex() for iteration.
    Those are not constrained to return only valid indexes (=isValidIndex() returns true).
    Therefore, users will most likely use findNextIndexWrap() etc. which only return valid
    object indexes, and can optionally filter for marked objects.

    If the underlying set changes (as opposed to: the underlying objects change),
    signal sig_set_changed is raised. */
class GObjectType {
 public:
    // Derived-class Interface:
    /** Virtual destructor. */
    virtual ~GObjectType();

    /** Get object, given an index.
        \param index Index
        \pre isValidIndex(index) */
    virtual GObject& getObjectByIndex(int index) = 0;

    /** Get universe, given an index.
        Note that the universe can be null.
        \param index
        \pre isValidIndex(index) */
    virtual GUniverse* getUniverseByIndex(int index) = 0;

    /** Check validity of an index.
        \param index Index, unconstrained
        \retval true object exists and can be accessed with getObjectByIndex()/getUniverseByIndex()
        \retval false object does not exist */
    virtual bool isValidIndex(int index) const = 0;

    /** Get next index.
        The returned index need not be valid as per isValidIndex(),
        but the implementation must guarantee that repeated calls to getNextIndex()
        ultimately end up at 0, so that loops actually terminate.
        There is no requirement that the indexes are reported in a particular order, though.
        \param index Starting index. Can be 0 to obtain the first index.
        \return 0 if no more objects, otherwise an index */
    virtual int getNextIndex(int index) const = 0;

    /** Get previous index.
        The returned index need not be valid as per isValidIndex(),
        but the implementation must guarantee that repeated calls to getNextIndex()
        ultimately end up at 0, so that loops actually terminate.
        There is no requirement that the indexes are reported in a particular order, though.
        \param index Starting index. Can be 0 to obtain the last index.
        \return 0 if no more objects. Otherwise an index */
    virtual int getPreviousIndex(int index) const = 0;

    // Derived:
    bool isEmpty() const;
    bool isUnit() const;
    int countObjects() const;
    int findNearestIndex(GPoint pt);
    int findPreviousIndexWrap(int index, bool marked);
    int findNextIndexWrap(int index, bool marked);
    int findPreviousIndexNoWrap(int index, bool marked);
    int findNextIndexNoWrap(int index, bool marked);
    int findNextIndex(int index) const;

    /** Called when the underlying set changes, i.e. objects come and go
        or are replaced by different objects. Called after the change.

        For simple changes, the integer can be a hint for users, i.e. the
        new Id of a renamed object. If the type doesn't want to give a hint,
        it can pass 0. */
    Signal1<int> sig_set_changed;
};

/*! \class GProxyObjectType
    \brief Proxy Object Type

    This is an Object Type that relays all requests to another, runtime-configurable
    object type. It can even refer to no object type at all, at which time it will
    behave as if it were empty. */
class GProxyObjectType : public GObjectType {
 public:
    GProxyObjectType();
    ~GProxyObjectType();

    // GObjectType:
    GObject& getObjectByIndex(int index);
    GUniverse* getUniverseByIndex(int index);
    bool isValidIndex(int index) const;
    int getNextIndex(int index) const;
    int getPreviousIndex(int index) const;

    // GProxyObjectType:
    void setObjectType(GObjectType* type);
    GObjectType* getObjectType();

 private:
    GObjectType* type;
    SignalConnection conn_set_changed;
};

/*! \class GIndexedObjectType
    \brief Indexed Object Type

    This class is intended to simplify implementation of descendants.
    It implements getNextIndex / getPreviousIndex as iteration through a set
    of natural numbers up to a maximum. Descendants only have to implement
    the actual object access. */
class GIndexedObjectType : public GObjectType {
 public:
    GIndexedObjectType(int max);
    ~GIndexedObjectType();

    // GObjectType:
    GObject& getObjectByIndex(int index) = 0;
    GUniverse* getUniverseByIndex(int index) = 0;
    bool isValidIndex(int index) const = 0;
    int getNextIndex(int index) const;
    int getPreviousIndex(int index) const;

 private:
    const int max;
};

#endif
